/* ----------------------------------------------------------------------------
 * rishu_proxy_response.cpp
 *   里珠/Proxy の回答の実装.
 * ----------------------------------------------------------------------------
 * Mastering programed by YAMASHINA Hio
 *
 * Copyright 2002 YMIRLINK,Inc.
 * ----------------------------------------------------------------------------
 * $Id: rishu_proxy_response.c,v 1.5 2002/11/10 13:53:18 hio Exp $
 * ----------------------------------------------------------------------------
 */

#include <windows.h>
#include "rishu_proxy.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/* ----------------------------------------------------------------------------
 * 初期化.
 */
RishuProxy_Response* RishuProxy_Response_init(RishuProxy_Response* response,
					      WORD wSize,
					      RishuProxy* proxy
					      )
{
  if( wSize<sizeof(RishuProxy_Response) )
  {
    return NULL;
  }
  response->wSize = wSize;
  response->wVersion = 0;
  response->proxy = proxy;
  /* */
  response->header = NULL;
  response->dwHeaderLen = 0;
  response->dwHeaderSize = 0;
  /* */
  response->body = NULL;
  response->dwBodyLen = 0;
  response->dwBodySize = 0;
  /* */
  response->statusLine = NULL;
  response->dwStatusLineLen = 0;
  response->nStatusCode = 0;
  return response;
}

/* ----------------------------------------------------------------------------
 * 破棄.
 */
void RishuProxy_Response_discard(RishuProxy_Response* response)
{
  BOOL proxyLocked = FALSE;
  if( response->header )
  {
    RishuProxy_lock(response->proxy);
    proxyLocked = TRUE;
    RishuProxy_async_myFree(response->proxy,response->header);
    response->header = NULL;
    response->dwHeaderLen = 0;
    response->dwHeaderSize = 0;
  }
  if( response->body )
  {
    if( !proxyLocked )
    {
      RishuProxy_lock(response->proxy);
      proxyLocked = TRUE;
    }
    RishuProxy_async_myFree(response->proxy,response->body);
    response->body = NULL;
    response->dwBodyLen = 0;
    response->dwBodySize = 0;
  }
  if( response->statusLine )
  {
    if( !proxyLocked )
    {
      RishuProxy_lock(response->proxy);
      proxyLocked = TRUE;
    }
    RishuProxy_async_myFree(response->proxy,response->statusLine);
    response->statusLine = NULL;
    response->dwStatusLineLen = 0;
  }
  if( proxyLocked )
  {
    RishuProxy_unlock(response->proxy);
  }
  response->proxy = NULL;
  
  return;
}

/* ----------------------------------------------------------------------------
 * 内容のクリア.
 */
void RishuProxy_Response_clear(RishuProxy_Response* response)
{
  response->wVersion = 0;
  response->dwHeaderLen = 0;
  response->dwBodyLen = 0;
  response->dwStatusLineLen = 0;
  response->nStatusCode = 0;  
  return;
}

/* ----------------------------------------------------------------------------
 * バージョンの取得
 */
WORD RishuProxy_Response_getVersion(RishuProxy_Response* response)
{
  if( response->wVersion )
  {
    return response->wVersion;
  }
  RishuProxy_Response_getStatusCode(response);
  return response->wVersion;
}

/* ----------------------------------------------------------------------------
 * ステータスコードの取得
 */
int RishuProxy_Response_getStatusCode(RishuProxy_Response* response)
{
  int nRes;
  int hi,low;
  int statusCode;
  
  if( response->nStatusCode )
  {
    return response->nStatusCode;
  }
  if( !response->statusLine )
  {
    RishuProxy_Response_getStatusLine(response);
    if( !response->statusLine )
    {
      return 0x0000;
    }
  }
  
  nRes = _snscanf((const char*)response->statusLine,
		  response->dwStatusLineLen,
		  "RISHU/%d.%d %d",
		  &hi, &low, &statusCode
		  );
  if( nRes!=3 )
  {
    return 0x0000;
  }
  response->wVersion = (hi<<8)|low;
  response->nStatusCode = statusCode;
  
  return statusCode;
}

/* ----------------------------------------------------------------------------
 * ステータス行の取得.
 */
LPVOID RishuProxy_Response_getStatusLine(RishuProxy_Response* response)
{
  DWORD i;
  
  if( !response->header )
  {
    return "";
  }
  if( response->statusLine )
  {
    return response->statusLine;
  }
  
  for( i=0; i<response->dwHeaderLen; ++i )
  {
    if( ((LPCBYTE)response->header)[i]=='\n' )
    {
      LPBYTE statusLine;
      if( i && ((LPCBYTE)response->header)[i-1]=='\r' )
      {
	--i;
      }
      statusLine = RishuProxy_myAlloc(response->proxy,i+1);
      if( !statusLine )
      {
	return "";
      }
      CopyMemory(statusLine,response->header,i);
      statusLine[i] = '\0';
      response->statusLine = statusLine;
      response->dwStatusLineLen = i;
      return response->statusLine;
    }
  }
  return "";
}

/* ----------------------------------------------------------------------------
 * HGLOBAL に変換.
 */
HGLOBAL RishuProxy_Response_intoGlobalMemory(RishuProxy_Response* response,
					     HGLOBAL hGlobal,
					     long* ptr_len)
{
  LPCBYTE ptr;
  const LPCBYTE tail = ((LPCBYTE)response->header)+response->dwHeaderLen;
  
  if( !response->header )
  {
    return RishuProxy_reply500(hGlobal,ptr_len);
  }
  
  ptr = response->header;
  /* ステータス行を解析. */
  RishuProxy_Response_getStatusLine(response);
  ptr += response->dwStatusLineLen;
  if( ptr >= tail ) return RishuProxy_reply500(hGlobal,ptr_len);
  if( *ptr=='\r' )
  {
    ++ptr;
    if( ptr >= tail ) return RishuProxy_reply500(hGlobal,ptr_len);
  }
  if( *ptr=='\n' )
  {
    ++ptr;
    if( ptr >= tail ) return RishuProxy_reply500(hGlobal,ptr_len);
  }
  
  /* Original-Response: の除去. */
  {
    BYTE cmp[] = "Original-Response:";
    DWORD cmplen = sizeof(cmp)-1;
    if( ptr+cmplen<tail )
    {
      if( strncmp(ptr,cmp,cmplen)==0 )
      {
        ptr += cmplen;
      }else
      { /* strncasecmp の代用. */
	int i;
	for( i=0; ; ++i )
	{
	  if( i==cmplen )
	  {
	    ptr += i;
	    break;
	  }
	  if( cmp[i]!=ptr[i] && tolower(cmp[i])!=tolower(ptr[i]) )
	  {
	    break;
	  }
	}
      }
    }
    while( isspace(*ptr) && ++ptr<tail );
  }
  
  /* HGLOBAL さいずちぇっく */
  if( tail-ptr>*ptr_len && (DWORD)(tail-ptr)>GlobalSize(hGlobal) )
  {
    const HGLOBAL hGlobal_bak = hGlobal;
#if 1
    hGlobal = GlobalFree(hGlobal);
    if( hGlobal!=NULL )
    {
      if( RishuProxy_errorlogIsEnabled(response->proxy) )
      {
	LPTSTR cause = getLastErrorMessage();
	RishuProxy_log(response->proxy, TEXT("error"),
		       TEXT("RishuProxy::Response::intoGlobalMemory"),
		       TEXT("[GlobalFree failed]\n"
			    "[[hGlobal==0x%08x]]\n"
			    "[[\n%s]]\n"),
		       hGlobal_bak,
		       cause
		       );
	LocalFree(cause);
      }
    }
    hGlobal = GlobalAlloc(GMEM_FIXED,tail-ptr);
    if( hGlobal==NULL )
    {
      if( RishuProxy_errorlogIsEnabled(response->proxy) )
      {
	LPTSTR cause = getLastErrorMessage();
	RishuProxy_log(response->proxy, TEXT("error"),
		       TEXT("RishuProxy::Response::intoGlobalMemory"),
		       TEXT("[GlobalAlloc failed]\n"
			    "[[size==%d]]\n"
			    "[[\n%s]]\n"),
		       tail-ptr,
		       cause
		       );
	LocalFree(cause);
      }
      return NULL;
    }
#else
    hGlobal = GlobalReAlloc(hGlobal,tail-ptr,0);
    if( !hGlobal )
    {
      if( RishuProxy_errorlogIsEnabled(response->proxy) )
      {
	LPTSTR cause = getLastErrorMessage();
	RishuProxy_log(response->proxy, TEXT("error"),
		       TEXT("RishuProxy::Response::intoGlobalMemory"),
		       TEXT("[GlobalReAlloc failed]\n"
			    "[[hGlobal==0x%08x]] [[size==%d]]\n"
			    "[[\n%s]]\n"),
		       hGlobal_bak,
		       tail-ptr,
		       cause
		       );
	LocalFree(cause);
      }
      *ptr_len = 0;
      return NULL;
    }
#endif
  }
  
  /* HGLOBAL にこぴー */
  *ptr_len = tail-ptr;
  CopyMemory((LPVOID)hGlobal,ptr,tail-ptr);
  
  if( RishuProxy_tracelogIsEnabled(response->proxy) )
  {
    RishuProxy_log(response->proxy, TEXT("trace"),
		   TEXT("RishuProxy::Response::intoGlobalMemory"),
		   TEXT("[leave]\n[[*ptr_len==%d]]\n[[\n%.*s]]\n"),
		   *ptr_len,
		   *ptr_len,hGlobal
		   );
  }
  
  return hGlobal;
}

/* ----------------------------------------------------------------------------
 * データの追加
 */
void RishuProxy_Response_write(RishuProxy_Response* response,
			       LPCVOID data,
			       DWORD data_len
			       )
{
  if( response->dwHeaderSize-response->dwHeaderLen<data_len )
  {
    if( !response->header )
    {
      response->header = RishuProxy_myAlloc(response->proxy,data_len);
      if( !response->header )
      {
	return;
      }
    }else
    {
      register const HGLOBAL olddata = response->header;
      response->header = RishuProxy_myReAlloc(response->proxy,olddata,response->dwHeaderLen+data_len);
      if( !response->header )
      {
	response->header = olddata;
	return;
      }
    }
    response->dwHeaderSize = RishuProxy_mySize(response->proxy,response->header);
  }
  CopyMemory((LPBYTE)response->header+response->dwHeaderLen,data,data_len);
  response->dwHeaderLen   += data_len;
  return;
}

